"""intro.py - a context for displaying a splash screen, slideshow, movie, etc.
"""

import random

import pygame

import clock
import context
from interpolation import Interpolator
import mainmenu
import pygametext
import settings


__version__ = '3.0.0'
__author__ = 'Gummbum, (c) 2017'
__vernum__ = tuple(int(s) for s in __version__.split('.'))


class Intro(context.Context):
    def __init__(self):
        context.Context.__init__(self, clock.FixedDriver(self.update, settings.time_step))
        self.clock.new_schedule(self.draw, settings.draw_step)

        self.bgcolor = pygame.Color('darkgreen')

        self.font = None
        self.countdown = None
        self.elapsed = None
        self.text = None
        self.text_img = None
        self.text_rect = None
        self.dx = None
        self.dy = None
        self.surfs = []
        self.interp_text = None

    def enter(self):
        # see context.Context
        font_info = settings.font_info.intro
        self.font = pygametext.getfont(fontname=font_info.fontname, fontsize=font_info.fontsize)
        self.elapsed = self.countdown = 10.99999  # float, seconds
        self.text = ''
        self.text_rect = pygame.Rect(3, 3, 1, 1)
        # pre-cache; generating this big text is very expensive
        for i in range(11):
            text = 'Intro (Wait {0:d} or Click)'.format(i)
            self.surfs.append(pygametext.getsurf(text, **settings.font_themes['intro']))
        self.update_text()
        self.dx = random.choice((-6, -5, -4, 4, 5, 6))
        self.dy = random.choice((-6, -5, -4, 4, 5, 6))
        self.interp_text = Interpolator(*self.text_rect.topleft)

    def exit(self):
        # see context.Context
        pass

    def suspend(self):
        # see context.Context
        pass

    def resume(self):
        # see context.Context
        pass

    def quit_context(self):
        self.pop()

    def update(self, dt):
        self.countdown -= dt
        if int(self.countdown) <= 0:
            self.action_main_menu()
            return
        
        self.get_events()
        self.update_text()
        text_rect = self.text_rect
        text_rect.x += self.dx
        text_rect.y += self.dy
        screen_rect = settings.screen_rect
        if not screen_rect.contains(text_rect):
            if text_rect.x < 0 or text_rect.right >= screen_rect.right:
                self.dx *= -1
            if text_rect.y < 0 or text_rect.bottom >= screen_rect.bottom:
                self.dy *= -1
            self.text_rect.clamp_ip(screen_rect)
        self.interp_text.update(*text_rect.topleft)

    def update_text(self):
        countdown = int(self.countdown)
        if countdown == self.elapsed:
            return
        self.elapsed = countdown
        self.text_img = self.surfs[countdown]
        self.text_rect.size = self.text_img.get_size()
    
    def draw(self, sched):
        screen = settings.screen
        screen.fill(self.bgcolor)
        ix, iy = self.interp_text.raw_offset(sched.interp)
        screen.blit(self.text_img, self.text_rect.move(ix, iy))
        pygame.display.flip()
    
    def action_main_menu(self):
        self.quit_context()
        self.push(mainmenu.MainMenu())

    def on_mouse_button_down(self, pos, button):
        # see context.Context.get_events()
        self.action_main_menu()

    def on_quit(self):
        # see context.Context.get_events()
        self.quit_context()
