"""settings.py - startup and runtime settings file

Usage
=====
Settings in this file should always be accessed like so:
    
    import settings
    if settings.whatever: ...
    settings.whatever = value

This insures that runtime changes will be noticed by other modules that import
the settings module. Of course, one can always make local copies of the values
if static behavior is desired.

See the end of the module for a section on custom environments for developers. This lets you set up your default
settings for distribution, and then override things for ease of development and debugging, without having to put all
the default settings back when you're done.
"""

import os

import pygame

import imagecache
import pygametext


__version__ = '3.0.0'
__author__ = 'Gummbum, (c) 2017'
__vernum__ = tuple(int(s) for s in __version__.split('.'))


class DictToSettings(object):
    """convert an arbitrarily nested dict to an object of nested attributes with the same hierarchy
    
    Caveat: does not work if the final value is a dict. :) In such a case do:
    
        thing = Setting(**dict(key1=dict(key2=None)))
        thing.key1.key2 = dict(...)
    """
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            if isinstance(value, dict):
                value = DictToSettings(**value)
            setattr(self, key, value)


class Setting(object):
    """easily build an object with nested attributes
    thing = Setting('key1.key2.key3', 123)
    thing = Setting()
    thing.key1.key2.key3 = 123
    print(thing.key1.key2.key3)
    """
    def __init__(self, attr=None, value=None):
        if isinstance(attr, str):
            setattr(self, attr, value)

    def __setattr__(self, attr, value):
        attrs = attr.split('.')
        if len(attrs) == 1:
            self.__dict__[attrs[0]] = value
        elif len(attrs) > 1:
            self.__dict__[attrs[0]] = Setting('.'.join(attrs[1:]), value)

    def __getattr__(self, attr):
        if attr in self.__dict__:
            return self.__dict__[attr]
        else:
            self.__dict__[attr] = Setting()
            return self.__dict__[attr]


# CPU
max_tps = (30, 60, 90)[0]                   # game loop, physics fixed step
max_fps = (0, 50, 100, 150, 200, 250)[3]    # decoupled blitting: multiples of 50 are smoother
time_step = 1.0 / max_tps
draw_step = 1.0 / max_fps if max_fps else 0.0


# Display

resolution = 800, 600               # screen size in pixels: width, height
display_flags = (0, pygame.DOUBLEBUF | pygame.FULLSCREEN)[0]
default_caption = 'Game Skel'       # pygame.display.set_caption()
default_icon = None                 # TODO
screen = None                       # set me: when screen is opened or changed (main.py does this once)
screen_rect = None                  # set me: when screen is opened or changed (main.py does this once)


# Graphics

image_dir = 'data/image'
pygametext.FONT_NAME_TEMPLATE = 'data/font/%s'

# cached surfaces (required by spritefx)
# Note: pygametext does its own caching. No need to cache surfaces from pygametext.getsurf() unless you transform them.
# Note: these settings are not tunable on the fly. If you want to change the image cache settings during runtime, you'll
# have to call the image_cache instance's methods.
# Note: If you rely on aging or memory cap, remember to call image_cache.update(), or purge_memory() or purge_old().
image_cache_expire = 1 * 60          # expire unused textures after N seconds; recommend 1 min (1 * 60)
image_cache_memory = 5 * 2 ** 20     # force expiration if memory is exceeded; recommend 5 MB (5 * 2**20)
image_cache = imagecache.ImageCache(image_cache_expire, image_cache_memory)
image_cache.enable_age_cap(True)

# font_themes is used with the functions in module pygametext
# Example call to render some text using a theme:
#   image = pygametext.getsurf('some text', **settings.font_themes['intro'])
# font_themes can be accessed by nested dict keys:
#   font_theme['intro']['fontname']
font_themes = dict(
    intro=dict(
        fontname='Boogaloo.ttf',
        fontsize=52,
        color='red2',
        gcolor='deepskyblue',
        ocolor='yellow',
        owidth=0.5,
        scolor=None,
        shadow=None,
    ),
    mainmenu=dict(
        fontname='Bubblegum_Sans.ttf',
        fontsize=80,
        color='red2',
        gcolor='orange1',
        ocolor='yellow',
        owidth=0.5,
        scolor=None,
        shadow=None,
    ),
    game=dict(
        fontname='Vera.ttf',
        fontsize=38,
        color='lightblue',
        gcolor='green4',
        ocolor=None,
        owidth=0.5,
        scolor='grey20',
        shadow=(-1, 1),
    ),
    floatingtext=dict(
        fontname='Vera.ttf',
        fontsize=18,
        color='red',
        gcolor='lightblue4',
        ocolor='yellow',
        owidth=0.5,
        scolor=None,
        shadow=None,
    ),
)
# font_info can be accessed by nested attributes:
#   font_info.intro.fontname
font_info = DictToSettings(**font_themes)


# Sound

sound_dir = 'data/sound'
music_dir = 'data/music'


# Debug

profile = False             # run cProfile if True


# Custom

os.environ['SDL_VIDEO_CENTERED'] = '1'
# See main.py for the effect of bypass_menu:
bypass_menu = None

try:
    # Custom development preferences
    # 1. Create or touch an empty file in your project's root dir, where run_game.py is located.
    # 2. Add a tuple('Who', 'touched_file') in the local_profiles sequence.
    # 3. Remove or add a case for your 'who' value.
    # 4. Important: DO NOT DISTRIBUTE your local touched_file. Exclude it from commits, etc.
    local_profiles = (
        # who   touched_file
        # ----- ------------
        ('Gumm', 'iamgumm'),
        ('DR0ID', 'iamdr0id'),
    )
    who = [name for name, touched_file in local_profiles if os.access(touched_file, os.F_OK)][0]
    if who == 'Gumm':
        print('## CUSTOM SETTINGS FOR Gumm (see settings.py)')
        bypass_menu = (None, 'game', 'credits')[0]
        map_size = (150, 150)
        profile = False
    del who, local_profiles
except Exception as e:
    # print(e)
    pass

del os, pygame, pygametext
